// zig fmt: off
const std = @import("std");
const cmp = @import("matchers/cmp.zig");

const archive = @import("matchers/archive.zig");
const audio = @import("matchers/audio.zig");
const bin = @import("matchers/bin.zig");
const book = @import("matchers/book.zig");
const doc = @import("matchers/doc.zig");
const font = @import("matchers/font.zig");
const image = @import("matchers/image.zig");
const odf = @import("matchers/odf.zig");
const text = @import("matchers/text.zig");
const video = @import("matchers/video.zig");

pub fn compareBytes(slice: []const u8, subSlice: []const u8, startOff: usize) bool {
  return cmp.compareBytes(slice, subSlice, startOff);
}

pub fn sniffMediaType(data: []const u8) ?[]const u8 {
  if (bin.isWasm(data)) return "application/wasm";
  if (bin.isExe(data) or bin.isDll(data)) return "application/x-msdownload";
  if (bin.isElf(data)) return "application/x-elf";
  if (bin.isJava(data)) return "application/x-java-applet";
  if (bin.isLlvm(data)) return "application/x-llvm";
  if (bin.isMach(data)) return "application/x-mach-binary";
  if (bin.isDex(data)) return "application/vnd.android.dex";
  if (bin.isDey(data)) return "application/vnd.android.dey";
  if (bin.isDer(data)) return "application/x-x509-ca-cert";
  if (bin.isCoff(data)) return "application/x-coff";
  if (bin.isPem(data)) return "application/x-pem-file";
  if (bin.isQcow2(data)) return "application/x-qemu-disk";

  if (audio.isMidi(data)) return "audio/midi";
  if (audio.isMp3(data)) return "audio/mpeg";
  if (audio.isM4a(data)) return "audio/mp4";
  if (audio.isOggOpus(data)) return "audio/ogg";
  if (audio.isOgg(data)) return "audio/ogg";
  if (audio.isFlac(data)) return "audio/flac";
  if (audio.isWav(data)) return "audio/wav";
  if (audio.isAmr(data)) return "audio/amr";
  if (audio.isAac(data)) return "audio/aac";
  if (audio.isAiff(data)) return "audio/aiff";
  if (audio.isDsf(data)) return "audio/dsf";
  if (audio.isApe(data)) return "audio/ape";

  if (book.isEpub(data)) return "application/epub+zip";
  if (book.isMobi(data)) return "application/x-mobipocket-ebook";

  if (doc.isDoc(data)) return "application/msword";
  if (doc.isDocx(data)) return "application/vnd.openxmlformats-officedocument.wordprocessingml.document";
  if (doc.isXls(data)) return "application/vnd.ms-excel";
  if (doc.isXlsx(data)) return "application/vnd.openxmlformats-officedocument.spreadsheetml.sheet";
  if (doc.isPpt(data)) return "application/vnd.ms-powerpoint";
  if (doc.isPptx(data)) return "application/vnd.openxmlformats-officedocument.presentationml.presentation";

  if (font.isWoff(data)) return "font/woff";
  if (font.isWoff2(data)) return "font/woff2";
  if (font.isTtf(data)) return "font/ttf";
  if (font.isOtf(data)) return "font/otf";

  if (image.isJpeg(data)) return "image/jpeg";
  if (image.isJpeg2000(data)) return "image/jp2";
  if (image.isPng(data)) return "image/png";
  if (image.isGif(data)) return "image/gif";
  if (image.isWebp(data)) return "image/webp";
  if (image.isCr2(data)) return "image/x-canon-cr2";
  if (image.isTiff(data)) return "image/tiff";
  if (image.isBmp(data)) return "image/bmp";
  if (image.isJxr(data)) return "image/jxr";
  if (image.isPsd(data)) return "image/vnd.adobe.photoshop";
  if (image.isIco(data)) return "image/x-icon";
  if (image.isJxl(data)) return "image/jxl";
  if (image.isHeif(data)) return "image/heif";
  if (image.isAvif(data)) return "image/avif";
  if (image.isOra(data)) return "image/openraster";

  if (odf.isOdt(data)) return "application/vnd.oasis.opendocument.text";
  if (odf.isOds(data)) return "application/vnd.oasis.opendocument.spreadsheet";
  if (odf.isOdp(data)) return "application/vnd.oasis.opendocument.presentation";

  if (archive.isZip(data)) return "application/zip";
  if (archive.isTar(data)) return "application/x-tar";
  if (archive.isRar(data)) return "application/x-rar-compressed";
  if (archive.isGz(data)) return "application/gzip";
  if (archive.isBz2(data)) return "application/x-bzip2";
  if (archive.is7z(data)) return "application/x-7z-compressed";
  if (archive.isPdf(data)) return "application/pdf";
  if (archive.isSwf(data)) return "application/x-shockwave-flash";
  if (archive.isRtf(data)) return "application/rtf";
  if (archive.isXz(data)) return "application/x-xz";
  if (archive.isSqlite(data)) return "application/x-sqlite3";
  if (archive.isZst(data)) return "application/zstd";
  if (archive.isLz4(data)) return "application/x-lz4";

  if (text.isHtml(data)) return "text/html";
  if (text.isXml(data)) return "application/xml";
  if (text.isShellscript(data)) return "text/x-shellscript";

  if (video.isMp4(data)) return "video/mp4";
  if (video.isM4v(data)) return "video/x-m4v";
  if (video.isMkv(data)) return "video/x-matroska";
  if (video.isWebm(data)) return "video/webm";
  if (video.isMov(data)) return "video/quicktime";
  if (video.isAvi(data)) return "video/x-msvideo";
  if (video.isWmv(data)) return "video/x-ms-wmv";
  if (video.isMpeg(data)) return "video/mpeg";
  if (video.isFlv(data)) return "video/x-flv";

  return null;
}

test "sniff media type" {
  // WASM
  const wasmData = [_]u8{ 0x00, 0x61, 0x73, 0x6D, 0x01, 0x00, 0x00, 0x00 };
  try std.testing.expectEqualStrings("application/wasm", sniffMediaType(&wasmData).?);

  // EXE
  const exeData = [_]u8{ 0x4D, 0x5A, 0x00 };
  try std.testing.expectEqualStrings("application/x-msdownload", sniffMediaType(&exeData).?);

  // ELF
  var elfData = [_]u8{0} ** 53;
  elfData[0] = 0x7F;
  elfData[1] = 0x45;
  elfData[2] = 0x4C;
  elfData[3] = 0x46;
  try std.testing.expectEqualStrings("application/x-elf", sniffMediaType(&elfData).?);

  // Java
  const javaData = [_]u8{ 0xCA, 0xFE, 0xBA, 0xBE, 0x00, 0x00, 0x00, 0x2D };
  try std.testing.expectEqualStrings("application/x-java-applet", sniffMediaType(&javaData).?);

  // LLVM
  const llvmData = [_]u8{ 0x42, 0x43, 0x00 };
  try std.testing.expectEqualStrings("application/x-llvm", sniffMediaType(&llvmData).?);

  // Mach
  const machData = [_]u8{ 0xCF, 0xFA, 0xED, 0xFE };
  try std.testing.expectEqualStrings("application/x-mach-binary", sniffMediaType(&machData).?);

  // DEX
  var dexData = [_]u8{0} ** 37;
  dexData[0] = 0x64;
  dexData[1] = 0x65;
  dexData[2] = 0x78;
  dexData[3] = 0x0A;
  dexData[36] = 0x70;
  try std.testing.expectEqualStrings("application/vnd.android.dex", sniffMediaType(&dexData).?);

  // DEY
  var deyData = [_]u8{0} ** 101;
  deyData[0] = 0x64;
  deyData[1] = 0x65;
  deyData[2] = 0x79;
  deyData[3] = 0x0A;
  deyData[40] = 0x64;
  deyData[41] = 0x65;
  deyData[42] = 0x78;
  deyData[43] = 0x0A;
  deyData[76] = 0x70;
  try std.testing.expectEqualStrings("application/vnd.android.dey", sniffMediaType(&deyData).?);

  // DER
  const derData = [_]u8{ 0x30, 0x82, 0x00 };
  try std.testing.expectEqualStrings("application/x-x509-ca-cert", sniffMediaType(&derData).?);

  // COFF
  const coffData = [_]u8{ 0x4C, 0x01, 0x00 };
  try std.testing.expectEqualStrings("application/x-coff", sniffMediaType(&coffData).?);

  // PEM
  const pemData = [_]u8{ '-', '-', '-', '-', '-', 'B', 'E', 'G', 'I', 'N', ' ', 'C' };
  try std.testing.expectEqualStrings("application/x-pem-file", sniffMediaType(&pemData).?);

  // QCOW2
  const qcow2Data = [_]u8{ 'Q', 'F', 'I', 0xFB, 0x00 };
  try std.testing.expectEqualStrings("application/x-qemu-disk", sniffMediaType(&qcow2Data).?);

  // ODT
  var odtBuf = [_]u8{0} ** 100;
  odtBuf[0] = 'P';
  odtBuf[1] = 'K';
  odtBuf[2] = 0x03;
  odtBuf[3] = 0x04;
  const mimetype = "mimetype";
  for (mimetype, 0..) |c, i| {
    odtBuf[0x1E + i] = c;
  }
  const text_mime = "vnd.oasis.opendocument.text";
  for (text_mime, 0..) |c, i| {
    odtBuf[0x32 + i] = c;
  }
  try std.testing.expectEqualStrings("application/vnd.oasis.opendocument.text", sniffMediaType(&odtBuf).?);

  // ODS
  var odsBuf = [_]u8{0} ** 100;
  odsBuf[0] = 'P';
  odsBuf[1] = 'K';
  odsBuf[2] = 0x03;
  odsBuf[3] = 0x04;
  for (mimetype, 0..) |c, i| {
    odsBuf[0x1E + i] = c;
  }
  const spreadsheet_mime = "vnd.oasis.opendocument.spreadsheet";
  for (spreadsheet_mime, 0..) |c, i| {
    odsBuf[0x32 + i] = c;
  }
  try std.testing.expectEqualStrings("application/vnd.oasis.opendocument.spreadsheet", sniffMediaType(&odsBuf).?);

  // ODP
  var odpBuf = [_]u8{0} ** 100;
  odpBuf[0] = 'P';
  odpBuf[1] = 'K';
  odpBuf[2] = 0x03;
  odpBuf[3] = 0x04;
  for (mimetype, 0..) |c, i| {
    odpBuf[0x1E + i] = c;
  }
  const presentation_mime = "vnd.oasis.opendocument.presentation";
  for (presentation_mime, 0..) |c, i| {
    odpBuf[0x32 + i] = c;
  }
  try std.testing.expectEqualStrings("application/vnd.oasis.opendocument.presentation", sniffMediaType(&odpBuf).?);

  // ePub
  var epubBuf = [_]u8{0} ** 58;
  epubBuf[0] = 0x50;
  epubBuf[1] = 0x4B;
  epubBuf[2] = 0x03;
  epubBuf[3] = 0x04;
  const epub_mimetype = "mimetypeapplication/epub+zip";
  for (epub_mimetype, 0..) |c, i| {
    epubBuf[30 + i] = c;
  }
  try std.testing.expectEqualStrings("application/epub+zip", sniffMediaType(&epubBuf).?);

  // MOBI
  var mobiBuf = [_]u8{0} ** 68;
  mobiBuf[60] = 0x42;
  mobiBuf[61] = 0x4F;
  mobiBuf[62] = 0x4F;
  mobiBuf[63] = 0x4B;
  mobiBuf[64] = 0x4D;
  mobiBuf[65] = 0x4F;
  mobiBuf[66] = 0x42;
  mobiBuf[67] = 0x49;
  try std.testing.expectEqualStrings("application/x-mobipocket-ebook", sniffMediaType(&mobiBuf).?);

  // HEIF
  var heifBuf = [_]u8{0} ** 20;
  heifBuf[0] = 0x00;
  heifBuf[1] = 0x00;
  heifBuf[2] = 0x00;
  heifBuf[3] = 0x14;
  heifBuf[4] = 'f';
  heifBuf[5] = 't';
  heifBuf[6] = 'y';
  heifBuf[7] = 'p';
  heifBuf[8] = 'h';
  heifBuf[9] = 'e';
  heifBuf[10] = 'i';
  heifBuf[11] = 'c';
  heifBuf[12] = 0x00;
  heifBuf[13] = 0x00;
  heifBuf[14] = 0x00;
  heifBuf[15] = 0x00;
  heifBuf[16] = 'h';
  heifBuf[17] = 'e';
  heifBuf[18] = 'i';
  heifBuf[19] = 'c';
  try std.testing.expectEqualStrings("image/heif", sniffMediaType(&heifBuf).?);

  // AVIF
  var avifBuf = [_]u8{0} ** 20;
  avifBuf[0] = 0x00;
  avifBuf[1] = 0x00;
  avifBuf[2] = 0x00;
  avifBuf[3] = 0x14;
  avifBuf[4] = 'f';
  avifBuf[5] = 't';
  avifBuf[6] = 'y';
  avifBuf[7] = 'p';
  avifBuf[8] = 'a';
  avifBuf[9] = 'v';
  avifBuf[10] = 'i';
  avifBuf[11] = 'f';
  avifBuf[12] = 0x00;
  avifBuf[13] = 0x00;
  avifBuf[14] = 0x00;
  avifBuf[15] = 0x00;
  avifBuf[16] = 'a';
  avifBuf[17] = 'v';
  avifBuf[18] = 'i';
  avifBuf[19] = 'f';
  try std.testing.expectEqualStrings("image/avif", sniffMediaType(&avifBuf).?);

  // ZIP
  const zipData = [_]u8{ 0x50, 0x4B, 0x03, 0x04 };
  try std.testing.expectEqualStrings("application/zip", sniffMediaType(&zipData).?);

  // TAR
  var tarData = [_]u8{0} ** 262;
  tarData[257] = 0x75;
  tarData[258] = 0x73;
  tarData[259] = 0x74;
  tarData[260] = 0x61;
  tarData[261] = 0x72;
  try std.testing.expectEqualStrings("application/x-tar", sniffMediaType(&tarData).?);

  // RAR
  const rarData = [_]u8{ 0x52, 0x61, 0x72, 0x21, 0x1A, 0x07, 0x00 };
  try std.testing.expectEqualStrings("application/x-rar-compressed", sniffMediaType(&rarData).?);

  // GZ
  const gzData = [_]u8{ 0x1F, 0x8B, 0x08 };
  try std.testing.expectEqualStrings("application/gzip", sniffMediaType(&gzData).?);

  // BZ2
  const bz2Data = [_]u8{ 0x42, 0x5A, 0x68 };
  try std.testing.expectEqualStrings("application/x-bzip2", sniffMediaType(&bz2Data).?);

  // 7Z
  const szData = [_]u8{ 0x37, 0x7A, 0xBC, 0xAF, 0x27, 0x1C };
  try std.testing.expectEqualStrings("application/x-7z-compressed", sniffMediaType(&szData).?);

  // PDF
  const pdfData = [_]u8{ 0x25, 0x50, 0x44, 0x46 };
  try std.testing.expectEqualStrings("application/pdf", sniffMediaType(&pdfData).?);

  // SWF
  const swfData = [_]u8{ 0x43, 0x57, 0x53 };
  try std.testing.expectEqualStrings("application/x-shockwave-flash", sniffMediaType(&swfData).?);

  // RTF
  const rtfData = [_]u8{ 0x7B, 0x5C, 0x72, 0x74, 0x66 };
  try std.testing.expectEqualStrings("application/rtf", sniffMediaType(&rtfData).?);

  // XZ
  const xzData = [_]u8{ 0xFD, 0x37, 0x7A, 0x58, 0x5A, 0x00 };
  try std.testing.expectEqualStrings("application/x-xz", sniffMediaType(&xzData).?);

  // SQLite
  const sqliteData = [_]u8{ 0x53, 0x51, 0x4C, 0x69 };
  try std.testing.expectEqualStrings("application/x-sqlite3", sniffMediaType(&sqliteData).?);

  // Zstd
  const zstData = [_]u8{ 0x28, 0xB5, 0x2F, 0xFD };
  try std.testing.expectEqualStrings("application/zstd", sniffMediaType(&zstData).?);

  // LZ4
  const lz4Data = [_]u8{ 0x04, 0x22, 0x4D, 0x18 };
  try std.testing.expectEqualStrings("application/x-lz4", sniffMediaType(&lz4Data).?);

  // DOC
  var docBuf = [_]u8{0} ** 96;
  docBuf[0] = 0xD0;
  docBuf[1] = 0xCF;
  docBuf[2] = 0x11;
  docBuf[3] = 0xE0;
  docBuf[4] = 0xA1;
  docBuf[5] = 0xB1;
  docBuf[6] = 0x1A;
  docBuf[7] = 0xE1;
  docBuf[80] = 0x06;
  docBuf[81] = 0x09;
  docBuf[82] = 0x02;
  docBuf[83] = 0x00;
  docBuf[84] = 0x00;
  docBuf[85] = 0x00;
  docBuf[86] = 0x00;
  docBuf[87] = 0x00;
  docBuf[88] = 0xC0;
  docBuf[89] = 0x00;
  docBuf[90] = 0x00;
  docBuf[91] = 0x00;
  docBuf[92] = 0x00;
  docBuf[93] = 0x00;
  docBuf[94] = 0x00;
  docBuf[95] = 0x46;
  try std.testing.expectEqualStrings("application/msword", sniffMediaType(&docBuf).?);

  // DOCX
  var docxBuf = [_]u8{0} ** 100;
  docxBuf[0] = 'P';
  docxBuf[1] = 'K';
  docxBuf[2] = 0x03;
  docxBuf[3] = 0x04;
  const word = "word/";
  for (word, 0..) |c, i| {
    docxBuf[0x1E + i] = c;
  }
  try std.testing.expectEqualStrings("application/vnd.openxmlformats-officedocument.wordprocessingml.document", sniffMediaType(&docxBuf).?);

  // XLS
  var xlsBuf = [_]u8{0} ** 96;
  xlsBuf[0] = 0xD0;
  xlsBuf[1] = 0xCF;
  xlsBuf[2] = 0x11;
  xlsBuf[3] = 0xE0;
  xlsBuf[4] = 0xA1;
  xlsBuf[5] = 0xB1;
  xlsBuf[6] = 0x1A;
  xlsBuf[7] = 0xE1;
  xlsBuf[80] = 0x10;
  xlsBuf[81] = 0x08;
  xlsBuf[82] = 0x02;
  xlsBuf[83] = 0x00;
  xlsBuf[84] = 0x00;
  xlsBuf[85] = 0x00;
  xlsBuf[86] = 0x00;
  xlsBuf[87] = 0x00;
  xlsBuf[88] = 0xC0;
  xlsBuf[89] = 0x00;
  xlsBuf[90] = 0x00;
  xlsBuf[91] = 0x00;
  xlsBuf[92] = 0x00;
  xlsBuf[93] = 0x00;
  xlsBuf[94] = 0x00;
  xlsBuf[95] = 0x46;
  try std.testing.expectEqualStrings("application/vnd.ms-excel", sniffMediaType(&xlsBuf).?);

  // XLSX
  var xlsxBuf = [_]u8{0} ** 100;
  xlsxBuf[0] = 'P';
  xlsxBuf[1] = 'K';
  xlsxBuf[2] = 0x03;
  xlsxBuf[3] = 0x04;
  const xl = "xl/";
  for (xl, 0..) |c, i| {
    xlsxBuf[0x1E + i] = c;
  }
  try std.testing.expectEqualStrings("application/vnd.openxmlformats-officedocument.spreadsheetml.sheet", sniffMediaType(&xlsxBuf).?);

  // PPT
  var pptBuf = [_]u8{0} ** 96;
  pptBuf[0] = 0xD0;
  pptBuf[1] = 0xCF;
  pptBuf[2] = 0x11;
  pptBuf[3] = 0xE0;
  pptBuf[4] = 0xA1;
  pptBuf[5] = 0xB1;
  pptBuf[6] = 0x1A;
  pptBuf[7] = 0xE1;
  pptBuf[80] = 0x10;
  pptBuf[81] = 0x8D;
  pptBuf[82] = 0x81;
  pptBuf[83] = 0x64;
  pptBuf[84] = 0x9B;
  pptBuf[85] = 0x4F;
  pptBuf[86] = 0xCF;
  pptBuf[87] = 0x11;
  pptBuf[88] = 0x86;
  pptBuf[89] = 0xEA;
  pptBuf[90] = 0x00;
  pptBuf[91] = 0xAA;
  pptBuf[92] = 0x00;
  pptBuf[93] = 0xB9;
  pptBuf[94] = 0x29;
  pptBuf[95] = 0xE8;
  try std.testing.expectEqualStrings("application/vnd.ms-powerpoint", sniffMediaType(&pptBuf).?);

  // PPTX
  var pptxBuf = [_]u8{0} ** 100;
  pptxBuf[0] = 'P';
  pptxBuf[1] = 'K';
  pptxBuf[2] = 0x03;
  pptxBuf[3] = 0x04;
  const ppt = "ppt/";
  for (ppt, 0..) |c, i| {
    pptxBuf[0x1E + i] = c;
  }
  try std.testing.expectEqualStrings("application/vnd.openxmlformats-officedocument.presentationml.presentation", sniffMediaType(&pptxBuf).?);

  // HTML
  const htmlData = "<HTML >";
  try std.testing.expectEqualStrings("text/html", sniffMediaType(htmlData).?);

  // XML
  const xmlData = "<?xml version=\"1.0\"?>";
  try std.testing.expectEqualStrings("application/xml", sniffMediaType(xmlData).?);

  // Shell script
  const shellData = "#!/bin/sh";
  try std.testing.expectEqualStrings("text/x-shellscript", sniffMediaType(shellData).?);

  const midiBuf = [_]u8{ 0x4D, 0x54, 0x68, 0x64 };
  try std.testing.expectEqualStrings("audio/midi", sniffMediaType(&midiBuf).?);

  const mp3Buf = [_]u8{ 0x49, 0x44, 0x33 };
  try std.testing.expectEqualStrings("audio/mpeg", sniffMediaType(&mp3Buf).?);

  var m4aBuf = [_]u8{0} ** 11;
  m4aBuf[4] = 0x66;
  m4aBuf[5] = 0x74;
  m4aBuf[6] = 0x79;
  m4aBuf[7] = 0x70;
  m4aBuf[8] = 0x4D;
  m4aBuf[9] = 0x34;
  m4aBuf[10] = 0x41;
  try std.testing.expectEqualStrings("audio/mp4", sniffMediaType(&m4aBuf).?);

  var opusBuf = [_]u8{0} ** 36;
  opusBuf[0] = 0x4F;
  opusBuf[1] = 0x67;
  opusBuf[2] = 0x67;
  opusBuf[3] = 0x53;
  opusBuf[28] = 0x4F;
  opusBuf[29] = 0x70;
  opusBuf[30] = 0x75;
  opusBuf[31] = 0x73;
  opusBuf[32] = 0x48;
  opusBuf[33] = 0x65;
  opusBuf[34] = 0x61;
  opusBuf[35] = 0x64;
  try std.testing.expectEqualStrings("audio/ogg", sniffMediaType(&opusBuf).?);

  const woffBuf = [_]u8{ 0x77, 0x4F, 0x46, 0x46, 0x00, 0x01, 0x00, 0x00 };
  try std.testing.expectEqualStrings("font/woff", sniffMediaType(&woffBuf).?);

  const jpegBuf = [_]u8{ 0xFF, 0xD8, 0xFF };
  try std.testing.expectEqualStrings("image/jpeg", sniffMediaType(&jpegBuf).?);

  var mp4Buf = [_]u8{0} ** 12;
  mp4Buf[4] = 'f';
  mp4Buf[5] = 't';
  mp4Buf[6] = 'y';
  mp4Buf[7] = 'p';
  mp4Buf[8] = 'i';
  mp4Buf[9] = 's';
  mp4Buf[10] = 'o';
  mp4Buf[11] = 'm';
  try std.testing.expectEqualStrings("video/mp4", sniffMediaType(&mp4Buf).?);

  // 不明の場合
  const unknown = [_]u8{ 0xFF, 0xFF, 0xFF };
  try std.testing.expectEqual(null, sniffMediaType(&unknown));
}