// zig fmt: off
const std = @import("std");

/// 与えられたバッファがMIDIデータであるかどうかを返します。
pub fn isMidi(buf: []const u8) bool {
  return buf.len > 3 and
    buf[0] == 0x4D and
    buf[1] == 0x54 and
    buf[2] == 0x68 and
    buf[3] == 0x64;
}

/// 与えられたバッファがMP3データであるかどうかを返します。
pub fn isMp3(buf: []const u8) bool {
  return buf.len > 2 and
    ((buf[0] == 0x49 and buf[1] == 0x44 and buf[2] == 0x33) or
		 (buf[0] == 0xFF and buf[1] == 0xFB));
}

/// 与えられたバッファがM4Aデータであるかどうかを返します。
pub fn isM4a(buf: []const u8) bool {
  return (buf.len > 10 and
    buf[4] == 0x66 and
    buf[5] == 0x74 and
    buf[6] == 0x79 and
    buf[7] == 0x70 and
    buf[8] == 0x4D and
    buf[9] == 0x34 and
    buf[10] == 0x41) or
    (buf.len > 3 and
     buf[0] == 0x4D and
     buf[1] == 0x34 and
     buf[2] == 0x41 and
     buf[3] == 0x20);
}

/// 与えられたバッファがOGGデータであるかどうかを返します。
pub fn isOgg(buf: []const u8) bool {
  return buf.len > 3 and
    buf[0] == 0x4F and
    buf[1] == 0x67 and
    buf[2] == 0x67 and
    buf[3] == 0x53;
}

/// 与えられたバッファがOGG Opusデータであるかどうかを返します。
pub fn isOggOpus(buf: []const u8) bool {
  if (!isOgg(buf)) return false;

  return buf.len > 35 and
    buf[28] == 0x4F and
    buf[29] == 0x70 and
    buf[30] == 0x75 and
    buf[31] == 0x73 and
    buf[32] == 0x48 and
    buf[33] == 0x65 and
    buf[34] == 0x61 and
    buf[35] == 0x64;
}

/// 与えられたバッファがFLACデータであるかどうかを返します。
pub fn isFlac(buf: []const u8) bool {
  return buf.len > 3 and
    buf[0] == 0x66 and
    buf[1] == 0x4C and
    buf[2] == 0x61 and
    buf[3] == 0x43;
}

/// 与えられたバッファがWAVデータであるかどうかを返します。
pub fn isWav(buf: []const u8) bool {
  return buf.len > 11 and
    buf[0] == 0x52 and
    buf[1] == 0x49 and
    buf[2] == 0x46 and
    buf[3] == 0x46 and
    buf[8] == 0x57 and
    buf[9] == 0x41 and
    buf[10] == 0x56 and
    buf[11] == 0x45;
}

/// 与えられたバッファがAMRデータであるかどうかを返します。
pub fn isAmr(buf: []const u8) bool {
  return buf.len > 5 and
    buf[0] == 0x23 and
    buf[1] == 0x21 and
    buf[2] == 0x41 and
    buf[3] == 0x4D and
    buf[4] == 0x52 and
    buf[5] == 0x0A;
}

/// 与えられたバッファがAACデータであるかどうかを返します。
pub fn isAac(buf: []const u8) bool {
  return buf.len > 1 and
    buf[0] == 0xFF and
    (buf[1] == 0xF1 or buf[1] == 0xF9);
}

/// 与えられたバッファがAIFFデータであるかどうかを返します。
pub fn isAiff(buf: []const u8) bool {
  return buf.len > 11 and
    buf[0] == 0x46 and
    buf[1] == 0x4F and
    buf[2] == 0x52 and
    buf[3] == 0x4D and
    buf[8] == 0x41 and
    buf[9] == 0x49 and
    buf[10] == 0x46 and
    buf[11] == 0x46;
}

/// 与えられたバッファがDSFデータであるかどうかを返します。
pub fn isDsf(buf: []const u8) bool {
  // ref: https://dsd-guide.com/sites/default/files/white-papers/DSFFileFormatSpec_E.pdf
  return buf.len > 4 and
    buf[0] == 'D' and
    buf[1] == 'S' and
    buf[2] == 'D' and
    buf[3] == ' ';
}

/// 与えられたバッファがAPE (Monkey's Audio)データであるかどうかを返します。
pub fn isApe(buf: []const u8) bool {
  // ref: https://github.com/fernandotcl/monkeys-audio/blob/master/src/MACLib/APEHeader.h
  return buf.len > 4 and
    buf[0] == 'M' and
    buf[1] == 'A' and
    buf[2] == 'C' and
    buf[3] == ' ';
}

test "is midi" {
  const midiBuf = [_]u8{ 0x4D, 0x54, 0x68, 0x64, 0x00 };
  try std.testing.expect(isMidi(&midiBuf));

  const invalid = [_]u8{ 0x4D, 0x54, 0x68, 0xFF };
  try std.testing.expect(!isMidi(&invalid));

  const short = [_]u8{ 0x4D, 0x54, 0x68 };
  try std.testing.expect(!isMidi(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isMidi(&empty));
}

test "is mp3" {
  const mp3_id3Buf = [_]u8{ 0x49, 0x44, 0x33, 0x00 };
  try std.testing.expect(isMp3(&mp3_id3Buf));

  const mp3_syncBuf = [_]u8{ 0xFF, 0xFB, 0x00 };
  try std.testing.expect(isMp3(&mp3_syncBuf));

  const invalid = [_]u8{ 0x49, 0x44, 0xFF };
  try std.testing.expect(!isMp3(&invalid));

  const short = [_]u8{ 0x49, 0x44 };
  try std.testing.expect(!isMp3(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isMp3(&empty));
}

test "is m4a" {
  var m4a_ftypBuf = [_]u8{0} ** 11;
  m4a_ftypBuf[4] = 0x66;
  m4a_ftypBuf[5] = 0x74;
  m4a_ftypBuf[6] = 0x79;
  m4a_ftypBuf[7] = 0x70;
  m4a_ftypBuf[8] = 0x4D;
  m4a_ftypBuf[9] = 0x34;
  m4a_ftypBuf[10] = 0x41;
  try std.testing.expect(isM4a(&m4a_ftypBuf));

  const m4a_sigBuf = [_]u8{ 0x4D, 0x34, 0x41, 0x20, 0x00 };
  try std.testing.expect(isM4a(&m4a_sigBuf));

  var invalid = m4a_ftypBuf;
  invalid[8] = 0xFF;
  try std.testing.expect(!isM4a(&invalid));

  const short = [_]u8{ 0x4D, 0x34, 0x41 };
  try std.testing.expect(!isM4a(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isM4a(&empty));
}

test "is ogg" {
  const oggBuf = [_]u8{ 0x4F, 0x67, 0x67, 0x53, 0x00 };
  try std.testing.expect(isOgg(&oggBuf));

  const invalid = [_]u8{ 0x4F, 0x67, 0x67, 0xFF };
  try std.testing.expect(!isOgg(&invalid));

  const short = [_]u8{ 0x4F, 0x67, 0x67 };
  try std.testing.expect(!isOgg(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isOgg(&empty));
}

test "is ogg opus" {
  var opusBuf = [_]u8{0} ** 36;
  opusBuf[0] = 0x4F;
  opusBuf[1] = 0x67;
  opusBuf[2] = 0x67;
  opusBuf[3] = 0x53;
  opusBuf[28] = 0x4F;
  opusBuf[29] = 0x70;
  opusBuf[30] = 0x75;
  opusBuf[31] = 0x73;
  opusBuf[32] = 0x48;
  opusBuf[33] = 0x65;
  opusBuf[34] = 0x61;
  opusBuf[35] = 0x64;
  try std.testing.expect(isOggOpus(&opusBuf));

  var invalid = opusBuf;
  invalid[28] = 0xFF;
  try std.testing.expect(!isOggOpus(&invalid));

  var no_oggBuf = opusBuf;
  no_oggBuf[0] = 0xFF;
  try std.testing.expect(!isOggOpus(&no_oggBuf));

  const short = [_]u8{ 0x4F, 0x67, 0x67, 0x53 };
  try std.testing.expect(!isOggOpus(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isOggOpus(&empty));
}

test "is flac" {
  const flacBuf = [_]u8{ 0x66, 0x4C, 0x61, 0x43, 0x00 };
  try std.testing.expect(isFlac(&flacBuf));

  const invalid = [_]u8{ 0x66, 0x4C, 0x61, 0xFF };
  try std.testing.expect(!isFlac(&invalid));

  const short = [_]u8{ 0x66, 0x4C, 0x61 };
  try std.testing.expect(!isFlac(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isFlac(&empty));
}

test "is wav" {
  var wavBuf = [_]u8{0} ** 12;
  wavBuf[0] = 0x52;
  wavBuf[1] = 0x49;
  wavBuf[2] = 0x46;
  wavBuf[3] = 0x46;
  wavBuf[8] = 0x57;
  wavBuf[9] = 0x41;
  wavBuf[10] = 0x56;
  wavBuf[11] = 0x45;
  try std.testing.expect(isWav(&wavBuf));

  var invalid = wavBuf;
  invalid[8] = 0xFF;
  try std.testing.expect(!isWav(&invalid));

  const short = [_]u8{ 0x52, 0x49, 0x46, 0x46 };
  try std.testing.expect(!isWav(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isWav(&empty));
}

test "is amr" {
  const amrBuf = [_]u8{ 0x23, 0x21, 0x41, 0x4D, 0x52, 0x0A, 0x00 };
  try std.testing.expect(isAmr(&amrBuf));

  const invalid = [_]u8{ 0x23, 0x21, 0x41, 0x4D, 0xFF, 0x0A };
  try std.testing.expect(!isAmr(&invalid));

  const short = [_]u8{ 0x23, 0x21, 0x41, 0x4D, 0x52 };
  try std.testing.expect(!isAmr(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isAmr(&empty));
}

test "is aac" {
  const aac_f1Buf = [_]u8{ 0xFF, 0xF1, 0x00 };
  try std.testing.expect(isAac(&aac_f1Buf));

  const aac_f9Buf = [_]u8{ 0xFF, 0xF9, 0x00 };
  try std.testing.expect(isAac(&aac_f9Buf));

  const invalid = [_]u8{ 0xFF, 0xFF, 0x00 };
  try std.testing.expect(!isAac(&invalid));

  const short = [_]u8{ 0xFF };
  try std.testing.expect(!isAac(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isAac(&empty));
}

test "is aiff" {
  var aiffBuf = [_]u8{0} ** 12;
  aiffBuf[0] = 0x46;
  aiffBuf[1] = 0x4F;
  aiffBuf[2] = 0x52;
  aiffBuf[3] = 0x4D;
  aiffBuf[8] = 0x41;
  aiffBuf[9] = 0x49;
  aiffBuf[10] = 0x46;
  aiffBuf[11] = 0x46;
  try std.testing.expect(isAiff(&aiffBuf));

  var invalid = aiffBuf;
  invalid[8] = 0xFF;
  try std.testing.expect(!isAiff(&invalid));

  const short = [_]u8{ 0x46, 0x4F, 0x52, 0x4D };
  try std.testing.expect(!isAiff(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isAiff(&empty));
}

test "is dsf" {
  const dsfBuf = [_]u8{ 'D', 'S', 'D', ' ', 0x00 };
  try std.testing.expect(isDsf(&dsfBuf));

  const invalid = [_]u8{ 'D', 'S', 'D', 0xFF };
  try std.testing.expect(!isDsf(&invalid));

  const short = [_]u8{ 'D', 'S', 'D' };
  try std.testing.expect(!isDsf(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isDsf(&empty));
}

test "is ape" {
  const apeBuf = [_]u8{ 'M', 'A', 'C', ' ', 0x00 };
  try std.testing.expect(isApe(&apeBuf));

  const invalid = [_]u8{ 'M', 'A', 'C', 0xFF };
  try std.testing.expect(!isApe(&invalid));

  const short = [_]u8{ 'M', 'A', 'C' };
  try std.testing.expect(!isApe(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isApe(&empty));
}