// zig fmt: off
const std = @import("std");

/// 与えられたバッファがWASMであるかどうかを返します。
pub fn isWasm(buf: []const u8) bool {
  return buf.len >= 8 and
    buf[0] == 0x00 and
    buf[1] == 0x61 and
    buf[2] == 0x73 and
    buf[3] == 0x6D and
    buf[4] == 0x01 and
    buf[5] == 0x00 and
    buf[6] == 0x00 and
    buf[7] == 0x00;
}

/// 与えられたバッファがEXEであるかどうかを返します。
pub fn isExe(buf: []const u8) bool {
  return buf.len > 1 and buf[0] == 0x4D and buf[1] == 0x5A;
}

/// 与えられたバッファがDLLであるかどうかを返します。
pub fn isDll(buf: []const u8) bool {
  return isExe(buf);
}

/// 与えられたバッファがELFであるかどうかを返します。
pub fn isElf(buf: []const u8) bool {
  return buf.len > 52 and
    buf[0] == 0x7F and
    buf[1] == 0x45 and
    buf[2] == 0x4C and
    buf[3] == 0x46;
}

/// 与えられたバッファがJavaバイトコードであるかどうかを返します。
pub fn isJava(buf: []const u8) bool {
  if (buf.len < 8 or
    buf[0] != 0xCA or
    buf[1] != 0xFE or
    buf[2] != 0xBA or
    buf[3] != 0xBE) return false;

  const minMaj = std.mem.readInt(u32, buf[4..8], .big);
  if (minMaj < 45) return false;

  return std.mem.readInt(u16, buf[6..8], .big) >= 45;
}

/// 与えられたバッファがLLVMビットコードであるかどうかを返します。
pub fn isLlvm(buf: []const u8) bool {
  return buf.len >= 2 and buf[0] == 0x42 and buf[1] == 0x43;
}

/// 与えられたバッファがMach-Oバイナリであるかどうかを返します。
pub fn isMach(buf: []const u8) bool {
  if (buf.len < 4) return false;

  if (buf[1] == 0xFA and
      buf[2] == 0xED and
      buf[3] == 0xFE and
      (buf[0] == 0xCF or buf[0] == 0xCE)) return true;

  if (buf[0] == 0xFE and
      buf[1] == 0xED and
      buf[2] == 0xFA and
      (buf[3] == 0xCF or buf[3] == 0xCE)) return true;

  if (buf[0] == 0xCA and
      buf[1] == 0xFE and
      buf[2] == 0xBA and
      buf[3] == 0xBE and
      buf.len >= 8) {
    return std.mem.readInt(u32, buf[4..8], .big) < 45;
  }

  return false;
}

/// 与えられたバッファがDalvik Executable (DEX)であるかどうかを返します。
pub fn isDex(buf: []const u8) bool {
  return buf.len > 36 and
    buf[0] == 0x64 and
    buf[1] == 0x65 and
    buf[2] == 0x78 and
    buf[3] == 0x0A and
    buf[36] == 0x70;
}

/// 与えられたバッファがDey Optimized Dalvik Executable (ODEX)であるかどうかを返します。
pub fn isDey(buf: []const u8) bool {
  return buf.len > 100 and
    buf[0] == 0x64 and
    buf[1] == 0x65 and
    buf[2] == 0x79 and
    buf[3] == 0x0A and
    isDex(buf[40..100]);
}

/// 与えられたバッファがDER暗号化されたX.509証明書であるかどうかを返します。
pub fn isDer(buf: []const u8) bool {
  return buf.len > 2 and buf[0] == 0x30 and buf[1] == 0x82;
}

/// 与えられたバッファがi386向けのCommon Object File Formatであるかどうかを返します。
pub fn isCoffI386(buf: []const u8) bool {
  return buf.len > 2 and buf[0] == 0x4C and buf[1] == 0x01;
}

/// 与えられたバッファがx64向けのCommon Object File Formatであるかどうかを返します。
pub fn isCoffX64(buf: []const u8) bool {
  return buf.len > 2 and buf[0] == 0x64 and buf[1] == 0x86;
}

/// 与えられたバッファがItanium向けのCommon Object File Formatであるかどうかを返します。
pub fn isCoffIa64(buf: []const u8) bool {
  return buf.len > 2 and buf[0] == 0x00 and buf[1] == 0x02;
}

/// 与えられたバッファがCommon Object File Formatであるかどうかを返します。
pub fn isCoff(buf: []const u8) bool {
  return isCoffX64(buf) or isCoffI386(buf) or isCoffIa64(buf);
}

/// 与えられたバッファがPEMであるかどうかを返します。
pub fn isPem(buf: []const u8) bool {
  return buf.len > 11 and
    buf[0] == '-' and
    buf[1] == '-' and
    buf[2] == '-' and
    buf[3] == '-' and
    buf[4] == '-' and
    buf[5] == 'B' and
    buf[6] == 'E' and
    buf[7] == 'G' and
    buf[8] == 'I' and
    buf[9] == 'N' and
    buf[10] == ' ';
}

/// 与えられたバッファがQCOW2ディスクであるかどうかを返します。
pub fn isQcow2(buf: []const u8) bool {
  return buf.len > 4 and
    buf[0] == 'Q' and
    buf[1] == 'F' and
    buf[2] == 'I' and
    buf[3] == 0xFB;
}

test "is wasm" {
  const wasmData = [_]u8{ 0x00, 0x61, 0x73, 0x6D, 0x01, 0x00, 0x00, 0x00 };
  try std.testing.expect(isWasm(&wasmData));

  const invalid = [_]u8{ 0x00, 0x62, 0x73, 0x6D };
  try std.testing.expect(!isWasm(&invalid));

  const short = [_]u8{ 0x00, 0x61, 0x73 };
  try std.testing.expect(!isWasm(&short));
}

test "is exe" {
  const exeData = [_]u8{ 0x4D, 0x5A, 0x00 };
  try std.testing.expect(isExe(&exeData));

  const invalid = [_]u8{ 0x4E, 0x5A };
  try std.testing.expect(!isExe(&invalid));

  const short = [_]u8{ 0x4D };
  try std.testing.expect(!isExe(&short));
}

test "is dll" {
  const dllData = [_]u8{ 0x4D, 0x5A, 0x00 };
  try std.testing.expect(isDll(&dllData));

  const invalid = [_]u8{ 0x4E, 0x5A };
  try std.testing.expect(!isDll(&invalid));
}

test "is elf" {
  var elfData = [_]u8{0} ** 53;
  elfData[0] = 0x7F;
  elfData[1] = 0x45;
  elfData[2] = 0x4C;
  elfData[3] = 0x46;
  try std.testing.expect(isElf(&elfData));

  const invalid = [_]u8{ 0x7F, 0x46, 0x4C, 0x46 };
  try std.testing.expect(!isElf(&invalid));

  const short = [_]u8{ 0x7F, 0x45, 0x4C, 0x46 };
  try std.testing.expect(!isElf(&short));
}

test "is java" {
  const javaData = [_]u8{ 0xCA, 0xFE, 0xBA, 0xBE, 0x00, 0x00, 0x00, 0x2D };
  try std.testing.expect(isJava(&javaData));

  const machData = [_]u8{ 0xCA, 0xFE, 0xBA, 0xBE, 0x00, 0x00, 0x00, 0x01 };
  try std.testing.expect(!isJava(&machData));

  const short = [_]u8{ 0xCA, 0xFE, 0xBA };
  try std.testing.expect(!isJava(&short));
}

test "is llvm" {
  const llvmData = [_]u8{ 0x42, 0x43, 0x00 };
  try std.testing.expect(isLlvm(&llvmData));

  const invalid = [_]u8{ 0x43, 0x43 };
  try std.testing.expect(!isLlvm(&invalid));
}

test "is mach" {
  const mach_x86 = [_]u8{ 0xCF, 0xFA, 0xED, 0xFE };
  try std.testing.expect(isMach(&mach_x86));

  const mach_fat = [_]u8{ 0xCA, 0xFE, 0xBA, 0xBE, 0x00, 0x00, 0x00, 0x01 };
  try std.testing.expect(isMach(&mach_fat));

  const invalid = [_]u8{ 0xCF, 0xFB, 0xED, 0xFE };
  try std.testing.expect(!isMach(&invalid));

  const short = [_]u8{ 0xCF };
  try std.testing.expect(!isMach(&short));
}

test "is dex" {
  var dexData = [_]u8{0} ** 37;
  dexData[0] = 0x64;
  dexData[1] = 0x65;
  dexData[2] = 0x78;
  dexData[3] = 0x0A;
  dexData[36] = 0x70;
  try std.testing.expect(isDex(&dexData));

  const invalid = [_]u8{ 0x64, 0x65, 0x79, 0x0A };
  try std.testing.expect(!isDex(&invalid));
}

test "is dey" {
  var deyData = [_]u8{0} ** 101;
  deyData[0] = 0x64;
  deyData[1] = 0x65;
  deyData[2] = 0x79;
  deyData[3] = 0x0A;
  deyData[40] = 0x64;
  deyData[41] = 0x65;
  deyData[42] = 0x78;
  deyData[43] = 0x0A;
  deyData[76] = 0x70;
  try std.testing.expect(isDey(&deyData));

  const invalid = [_]u8{0} ** 101;
  try std.testing.expect(!isDey(&invalid));
}

test "is der" {
  const derData = [_]u8{ 0x30, 0x82, 0x00 };
  try std.testing.expect(isDer(&derData));

  const invalid = [_]u8{ 0x31, 0x82 };
  try std.testing.expect(!isDer(&invalid));
}

test "is coff i386" {
  const coffData = [_]u8{ 0x4C, 0x01, 0x00 };
  try std.testing.expect(isCoffI386(&coffData));

  const invalid = [_]u8{ 0x4D, 0x01 };
  try std.testing.expect(!isCoffI386(&invalid));
}

test "is coff x64" {
  const coffData = [_]u8{ 0x64, 0x86, 0x00 };
  try std.testing.expect(isCoffX64(&coffData));

  const invalid = [_]u8{ 0x65, 0x86 };
  try std.testing.expect(!isCoffX64(&invalid));
}

test "is coff ia64" {
  const coffData = [_]u8{ 0x00, 0x02, 0x00 };
  try std.testing.expect(isCoffIa64(&coffData));

  const invalid = [_]u8{ 0x00, 0x03 };
  try std.testing.expect(!isCoffIa64(&invalid));
}

test "is coff" {
  const i386Data = [_]u8{ 0x4C, 0x01, 0x00 };
  const x64Data = [_]u8{ 0x64, 0x86, 0x00 };
  const ia64Data = [_]u8{ 0x00, 0x02, 0x00 };
  try std.testing.expect(isCoff(&i386Data));
  try std.testing.expect(isCoff(&x64Data));
  try std.testing.expect(isCoff(&ia64Data));

  const invalid = [_]u8{ 0xFF, 0xFF };
  try std.testing.expect(!isCoff(&invalid));
}

test "is pem" {
  const pemData = [_]u8{ '-', '-', '-', '-', '-', 'B', 'E', 'G', 'I', 'N', ' ', 'C' };
  try std.testing.expect(isPem(&pemData));

  const invalid = [_]u8{ '-', '-', '-', '-', 'X', 'B', 'E', 'G', 'I', 'N', ' ' };
  try std.testing.expect(!isPem(&invalid));

  const short = [_]u8{ '-', '-', '-' };
  try std.testing.expect(!isPem(&short));
}

test "is qcow2" {
  const qcow2Data = [_]u8{ 'Q', 'F', 'I', 0xFB, 0x00 };
  try std.testing.expect(isQcow2(&qcow2Data));

  const invalid = [_]u8{ 'Q', 'F', 'J', 0xFB };
  try std.testing.expect(!isQcow2(&invalid));
}