// zig fmt: off
const std = @import("std");
const m = @import("cmp.zig");

const DocType = enum {
  DOC,
  DOCX,
  XLS,
  XLSX,
  PPT,
  PPTX,
  OOXML,
};

/// 与えられたバッファがMicrosoft Word Document (DOC) データであるかどうかを返します。
pub fn isDoc(buf: []const u8) bool {
  return ole2(buf) == .DOC;
}

/// 与えられたバッファがMicrosoft Word Open XML Format Document (DOCX) データであるかどうかを返します。
pub fn isDocx(buf: []const u8) bool {
  return msooxml(buf) == .DOCX;
}

/// 与えられたバッファがMicrosoft Excel 97-2003 Worksheet (XLS) データであるかどうかを返します。
pub fn isXls(buf: []const u8) bool {
  return ole2(buf) == .XLS;
}

/// 与えられたバッファがMicrosoft Excel Open XML Format Spreadsheet (XLSX) データであるかどうかを返します。
pub fn isXlsx(buf: []const u8) bool {
  return msooxml(buf) == .XLSX;
}

/// 与えられたバッファがMicrosoft PowerPoint 97-2003 Presentation (PPT) データであるかどうかを返します。
pub fn isPpt(buf: []const u8) bool {
  return ole2(buf) == .PPT;
}

/// 与えられたバッファがMicrosoft PowerPoint Open XML Presentation (PPTX) データであるかどうかを返します。
pub fn isPptx(buf: []const u8) bool {
  return msooxml(buf) == .PPTX;
}

fn msooxml(buf: []const u8) ?DocType {
  const signature = [_]u8{ 'P', 'K', 0x03, 0x04 };
  if (!m.compareBytes(buf, &signature, 0)) return null;
  if (checkMsooxml(buf, 0x1E)) |v| return v;

  if (!m.compareBytes(buf, "[Content_Types].xml", 0x1E) and
      !m.compareBytes(buf, "_rels/.rels", 0x1E) and
      !m.compareBytes(buf, "docProps", 0x1E)) return null;

  if (buf.len < 22) return null;
  const fileNameLen = std.mem.readInt(u32, buf[18..22], .little);
  var startOffset = @as(usize, fileNameLen) + 49;

  if (search(buf, startOffset, 6000)) |idx| {
    startOffset += idx + 4 + 26;

    if (search(buf, startOffset, 6000)) |idx2| {
      startOffset += idx2 + 4 + 26;

      if (checkMsooxml(buf, startOffset)) |typo| return typo;

      startOffset += 26;
      if (search(buf, startOffset, 6000)) |idx3| {
        startOffset += idx3 + 4 + 26;
        if (checkMsooxml(buf, startOffset)) |typo| return typo;
      }

      return .OOXML;
    }
  }

  return null;
}

fn ole2(buf: []const u8) ?DocType {
  const signature = [_]u8{ 0xD0, 0xCF, 0x11, 0xE0, 0xA1, 0xB1, 0x1A, 0xE1 };
  if (!m.compareBytes(buf, &signature, 0)) return null;
  if (buf.len < 96) return .DOC;

  const clsid = buf[80..96];
  const xlsClsid1 = [_]u8{ 0x10, 0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 };
  const xlsClsid2 = [_]u8{ 0x20, 0x08, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 };
  const docClsid = [_]u8{ 0x06, 0x09, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00, 0xC0, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x46 };
  const pptClsid = [_]u8{ 0x10, 0x8D, 0x81, 0x64, 0x9B, 0x4F, 0xCF, 0x11, 0x86, 0xEA, 0x00, 0xAA, 0x00, 0xB9, 0x29, 0xE8 };

  if (std.mem.eql(u8, clsid, &xlsClsid1) or std.mem.eql(u8, clsid, &xlsClsid2)) {
    return .XLS;
  } else if (std.mem.eql(u8, clsid, &docClsid)) {
    return .DOC;
  } else if (std.mem.eql(u8, clsid, &pptClsid)) {
    return .PPT;
  }

  return .DOC;
}

fn checkMsooxml(buf: []const u8, offset: usize) ?DocType {
  if (m.compareBytes(buf, "word/", offset)) {
    return .DOCX;
  } else if (m.compareBytes(buf, "ppt/", offset)) {
    return .PPTX;
  } else if (m.compareBytes(buf, "xl/", offset)) {
    return .XLSX;
  }
  return null;
}

fn search(buf: []const u8, start: usize, range: usize) ?usize {
  const length = buf.len;
  const signature = [_]u8{ 'P', 'K', 0x03, 0x04 };

  var end = start + range;
  if (end > length) end = length;
  if (start >= end) return null;

  for (start..end - signature.len + 1, 0..) |i, j| {
    if (std.mem.eql(u8, buf[i..i + signature.len], &signature)) return j;
  }

  return null;
}

test "is doc" {
  var docBuf = [_]u8{0} ** 96;
  docBuf[0] = 0xD0;
  docBuf[1] = 0xCF;
  docBuf[2] = 0x11;
  docBuf[3] = 0xE0;
  docBuf[4] = 0xA1;
  docBuf[5] = 0xB1;
  docBuf[6] = 0x1A;
  docBuf[7] = 0xE1;
  docBuf[80] = 0x06;
  docBuf[81] = 0x09;
  docBuf[82] = 0x02;
  docBuf[83] = 0x00;
  docBuf[84] = 0x00;
  docBuf[85] = 0x00;
  docBuf[86] = 0x00;
  docBuf[87] = 0x00;
  docBuf[88] = 0xC0;
  docBuf[89] = 0x00;
  docBuf[90] = 0x00;
  docBuf[91] = 0x00;
  docBuf[92] = 0x00;
  docBuf[93] = 0x00;
  docBuf[94] = 0x00;
  docBuf[95] = 0x46;
  try std.testing.expect(isDoc(&docBuf));

  const invalid = [_]u8{ 0xD0, 0xCF, 0x11, 0xE1, 0xA1, 0xB1, 0x1A, 0xE1 };
  try std.testing.expect(!isDoc(&invalid));

  const empty = [_]u8{};
  try std.testing.expect(!isDoc(&empty));
}

test "is docx" {
  var docxBuf = [_]u8{0} ** 100;
  docxBuf[0] = 'P';
  docxBuf[1] = 'K';
  docxBuf[2] = 0x03;
  docxBuf[3] = 0x04;
  const word = "word/";
  for (word, 0..) |c, i| {
    docxBuf[0x1E + i] = c;
  }
  try std.testing.expect(isDocx(&docxBuf));

  var invalid = docxBuf;
  invalid[0] = 'X';
  try std.testing.expect(!isDocx(&invalid));

  const short = [_]u8{ 'P', 'K', 0x03 };
  try std.testing.expect(!isDocx(&short));
}

test "is xls" {
  var xlsBuf = [_]u8{0} ** 96;
  xlsBuf[0] = 0xD0;
  xlsBuf[1] = 0xCF;
  xlsBuf[2] = 0x11;
  xlsBuf[3] = 0xE0;
  xlsBuf[4] = 0xA1;
  xlsBuf[5] = 0xB1;
  xlsBuf[6] = 0x1A;
  xlsBuf[7] = 0xE1;
  xlsBuf[80] = 0x10;
  xlsBuf[81] = 0x08;
  xlsBuf[82] = 0x02;
  xlsBuf[83] = 0x00;
  xlsBuf[84] = 0x00;
  xlsBuf[85] = 0x00;
  xlsBuf[86] = 0x00;
  xlsBuf[87] = 0x00;
  xlsBuf[88] = 0xC0;
  xlsBuf[89] = 0x00;
  xlsBuf[90] = 0x00;
  xlsBuf[91] = 0x00;
  xlsBuf[92] = 0x00;
  xlsBuf[93] = 0x00;
  xlsBuf[94] = 0x00;
  xlsBuf[95] = 0x46;
  try std.testing.expect(isXls(&xlsBuf));

  const invalid = [_]u8{ 0xD0, 0xCF, 0x11, 0xE1, 0xA1, 0xB1, 0x1A, 0xE1 };
  try std.testing.expect(!isXls(&invalid));

  const empty = [_]u8{};
  try std.testing.expect(!isXls(&empty));
}

test "is xlsx" {
  var xlsxBuf = [_]u8{0} ** 100;
  xlsxBuf[0] = 'P';
  xlsxBuf[1] = 'K';
  xlsxBuf[2] = 0x03;
  xlsxBuf[3] = 0x04;
  const xl = "xl/";
  for (xl, 0..) |c, i| {
    xlsxBuf[0x1E + i] = c;
  }
  try std.testing.expect(isXlsx(&xlsxBuf));

  var invalid = xlsxBuf;
  invalid[0] = 'X';
  try std.testing.expect(!isXlsx(&invalid));

  const short = [_]u8{ 'P', 'K', 0x03 };
  try std.testing.expect(!isXlsx(&short));
}

test "is ppt" {
  var pptBuf = [_]u8{0} ** 96;
  pptBuf[0] = 0xD0;
  pptBuf[1] = 0xCF;
  pptBuf[2] = 0x11;
  pptBuf[3] = 0xE0;
  pptBuf[4] = 0xA1;
  pptBuf[5] = 0xB1;
  pptBuf[6] = 0x1A;
  pptBuf[7] = 0xE1;
  pptBuf[80] = 0x10;
  pptBuf[81] = 0x8D;
  pptBuf[82] = 0x81;
  pptBuf[83] = 0x64;
  pptBuf[84] = 0x9B;
  pptBuf[85] = 0x4F;
  pptBuf[86] = 0xCF;
  pptBuf[87] = 0x11;
  pptBuf[88] = 0x86;
  pptBuf[89] = 0xEA;
  pptBuf[90] = 0x00;
  pptBuf[91] = 0xAA;
  pptBuf[92] = 0x00;
  pptBuf[93] = 0xB9;
  pptBuf[94] = 0x29;
  pptBuf[95] = 0xE8;
  try std.testing.expect(isPpt(&pptBuf));

  const invalid = [_]u8{ 0xD0, 0xCF, 0x11, 0xE1, 0xA1, 0xB1, 0x1A, 0xE1 };
  try std.testing.expect(!isPpt(&invalid));

  const empty = [_]u8{};
  try std.testing.expect(!isPpt(&empty));
}

test "is pptx" {
  var pptxBuf = [_]u8{0} ** 100;
  pptxBuf[0] = 'P';
  pptxBuf[1] = 'K';
  pptxBuf[2] = 0x03;
  pptxBuf[3] = 0x04;
  const ppt = "ppt/";
  for (ppt, 0..) |c, i| {
    pptxBuf[0x1E + i] = c;
  }
  try std.testing.expect(isPptx(&pptxBuf));

  var invalid = pptxBuf;
  invalid[0] = 'X';
  try std.testing.expect(!isPptx(&invalid));

  const short = [_]u8{ 'P', 'K', 0x03 };
  try std.testing.expect(!isPptx(&short));
}