// zig fmt: off
const std = @import("std");

/// 与えられたバッファがJPEG画像データであるかどうかを返します。
pub fn isJpeg(buf: []const u8) bool {
  return buf.len > 2 and buf[0] == 0xFF and buf[1] == 0xD8 and buf[2] == 0xFF;
}

/// 与えられたバッファがJPG2画像データであるかどうかを返します。
pub fn isJpeg2000(buf: []const u8) bool {
  return buf.len > 12 and
    buf[0] == 0x00 and
    buf[1] == 0x00 and
    buf[2] == 0x00 and
    buf[3] == 0x0C and
    buf[4] == 0x6A and
    buf[5] == 0x50 and
    buf[6] == 0x20 and
    buf[7] == 0x20 and
    buf[8] == 0x0D and
    buf[9] == 0x0A and
    buf[10] == 0x87 and
    buf[11] == 0x0A and
    buf[12] == 0x00;
}

/// 与えられたバッファがPNG画像データであるかどうかを返します。
pub fn isPng(buf: []const u8) bool {
  return buf.len > 3 and
    buf[0] == 0x89 and
    buf[1] == 0x50 and
    buf[2] == 0x4E and
    buf[3] == 0x47;
}

/// 与えられたバッファがGIF画像データであるかどうかを返します。
pub fn isGif(buf: []const u8) bool {
  return buf.len > 2 and
    buf[0] == 0x47 and
    buf[1] == 0x49 and
    buf[2] == 0x46;
}

/// 与えられたバッファがWEBP画像データであるかどうかを返します。
pub fn isWebp(buf: []const u8) bool {
  return buf.len > 11 and
    buf[8] == 0x57 and
    buf[9] == 0x45 and
    buf[10] == 0x42 and
    buf[11] == 0x50;
}

/// 与えられたバッファがCanon CR2画像データであるかどうかを返します。
pub fn isCr2(buf: []const u8) bool {
  return buf.len > 10 and
    ((buf[0] == 0x49 and
      buf[1] == 0x49 and
      buf[2] == 0x2A and
      buf[3] == 0x00) or
     (buf[0] == 0x4D and
      buf[1] == 0x4D and
      buf[2] == 0x00 and
      buf[3] == 0x2A)) and
    buf[8] == 0x43 and
    buf[9] == 0x52 and
    buf[10] == 0x02;
}

/// 与えられたバッファがTIFF画像データであるかどうかを返します。
pub fn isTiff(buf: []const u8) bool {
  return buf.len > 9 and
    ((buf[0] == 0x49 and
      buf[1] == 0x49 and
      buf[2] == 0x2A and
      buf[3] == 0x00) or
     (buf[0] == 0x4D and
      buf[1] == 0x4D and
      buf[2] == 0x00 and
      buf[3] == 0x2A)) and
    buf[8] != 0x43 and
    buf[9] != 0x52 and
    !isCr2(buf);
}

/// 与えられたバッファがBMP画像データであるかどうかを返します。
pub fn isBmp(buf: []const u8) bool {
  return buf.len > 1 and
    buf[0] == 0x42 and
    buf[1] == 0x4D;
}

/// 与えられたバッファがJXR画像データであるかどうかを返します。
pub fn isJxr(buf: []const u8) bool {
  return buf.len > 2 and
    buf[0] == 0x49 and
    buf[1] == 0x49 and
    buf[2] == 0xBC;
}

/// 与えられたバッファがPhotoship PSD画像データであるかどうかを返します。
pub fn isPsd(buf: []const u8) bool {
  return buf.len > 3 and
    buf[0] == 0x38 and
    buf[1] == 0x42 and
    buf[2] == 0x50 and
    buf[3] == 0x53;
}

/// 与えられたバッファがICOアイコン画像データであるかどうかを返します。
pub fn isIco(buf: []const u8) bool {
  return buf.len > 3 and
    buf[0] == 0x00 and
    buf[1] == 0x00 and
    buf[2] == 0x01 and
    buf[3] == 0x00;
}

/// 与えられたバッファがJPEG XL (JXL)画像データであるかどうかを返します。
pub fn isJxl(buf: []const u8) bool {
  return (buf.len > 2 and
    buf[0] == 0xFF and
    buf[1] == 0x0A) or
   (buf.len > 12 and
    buf[0] == 0x00 and
    buf[1] == 0x00 and
    buf[2] == 0x00 and
    buf[3] == 0x0C and
    buf[4] == 0x4A and
    buf[5] == 0x58 and
    buf[6] == 0x4C and
    buf[7] == 0x20 and
    buf[8] == 0x0D and
    buf[9] == 0x0A and
    buf[10] == 0x87 and
    buf[11] == 0x0A);
}

/// 与えられたバッファがHEIF画像データであるかどうかを返します。
pub fn isHeif(buf: []const u8) bool {
  if (buf.len == 0) return false;
  if (!isIsobmff(buf)) return false;

  if (getFtyp(buf)) |ftyp| {
    const heic = [_]u8{ 'h', 'e', 'i', 'c' };
    const heix = [_]u8{ 'h', 'e', 'i', 'x' };
    const mif1 = [_]u8{ 'm', 'i', 'f', '1' };
    const msf1 = [_]u8{ 'm', 's', 'f', '1' };

    if (std.mem.eql(u8, ftyp.major, &heic) or std.mem.eql(u8, ftyp.major, &heix)) {
      return true;
    }

    if (std.mem.eql(u8, ftyp.major, &mif1) or std.mem.eql(u8, ftyp.major, &msf1)) {
      for (ftyp.compatible, 0..) |brand, i| {
        if (i % 4 == 0 and brand == 'h' and
            i + 1 < ftyp.compatible.len and ftyp.compatible[i + 1] == 'e' and
            i + 2 < ftyp.compatible.len and ftyp.compatible[i + 2] == 'i' and
            i + 3 < ftyp.compatible.len and ftyp.compatible[i + 3] == 'c') {
          return true;
        }
      }
    }
  }

  return false;
}

/// 与えられたバッファがAVIF画像データであるかどうかを返します。
pub fn isAvif(buf: []const u8) bool {
  if (buf.len == 0) return false;
  if (!isIsobmff(buf)) return false;

  if (getFtyp(buf)) |ftyp| {
    const avif = [_]u8{ 'a', 'v', 'i', 'f' };
    const avis = [_]u8{ 'a', 'v', 'i', 's' };

    if (std.mem.eql(u8, ftyp.major, &avif) or std.mem.eql(u8, ftyp.major, &avis)) {
      return true;
    }

    var i: usize = 0;
    while (i + 3 < ftyp.compatible.len) : (i += 4) {
      const brand = ftyp.compatible[i..][0..4];
      if (std.mem.eql(u8, brand, &avif) or std.mem.eql(u8, brand, &avis)) {
        return true;
      }
    }
  }

  return false;
}

fn isIsobmff(buf: []const u8) bool {
  if (buf.len < 16) return false;

  const box_type = [_]u8{ 'f', 't', 'y', 'p' };
  return std.mem.readInt(u32, buf[0..4], .big) >= 8 and
    std.mem.eql(u8, buf[4..8], &box_type);
}

/// 与えられたバッファがORA画像データであるかどうかを返します。
pub fn isOra(buf: []const u8) bool {
  return buf.len > 57 and
    buf[0] == 0x50 and
    buf[1] == 0x4B and
    buf[2] == 0x03 and
    buf[3] == 0x04 and
    buf[30] == 0x6D and
    buf[31] == 0x69 and
    buf[32] == 0x6D and
    buf[33] == 0x65 and
    buf[34] == 0x74 and
    buf[35] == 0x79 and
    buf[36] == 0x70 and
    buf[37] == 0x65 and
    buf[38] == 0x69 and
    buf[39] == 0x6D and
    buf[40] == 0x61 and
    buf[41] == 0x67 and
    buf[42] == 0x65 and
    buf[43] == 0x2F and
    buf[44] == 0x6F and
    buf[45] == 0x70 and
    buf[46] == 0x65 and
    buf[47] == 0x6E and
    buf[48] == 0x72 and
    buf[49] == 0x61 and
    buf[50] == 0x73 and
    buf[51] == 0x74 and
    buf[52] == 0x65 and
    buf[53] == 0x72;
}

/// 与えられたバッファがDjVu画像データであるかどうかを返します。
pub fn is_djvu(buf: []const u8) bool {
  return buf.len > 14 and
    buf[0] == 0x41 and
    buf[1] == 0x54 and
    buf[2] == 0x26 and
    buf[3] == 0x54 and
    buf[4] == 0x46 and
    buf[5] == 0x4F and
    buf[6] == 0x52 and
    buf[7] == 0x4D and
    buf[12] == 0x44 and
    buf[13] == 0x4A and
    buf[14] == 0x56;
}

const Ftyp = struct {
  major: []const u8,
  minor: u32,
  compatible: []const u8,
};

fn getFtyp(buf: []const u8) ?Ftyp {
  if (!isIsobmff(buf)) return null;
  if (buf.len < 16) return null;

  const boxLen = std.mem.readInt(u32, buf[0..4], .big);
  if (boxLen < 16 or @as(usize, boxLen) > buf.len) return null;

  return Ftyp{
    .major = buf[8..12],
    .minor = std.mem.readInt(u32, buf[12..16], .big),
    .compatible = buf[16..boxLen],
  };
}

test "is jpeg" {
  const jpegBuf = [_]u8{ 0xFF, 0xD8, 0xFF, 0x00 };
  try std.testing.expect(isJpeg(&jpegBuf));

  const invalid = [_]u8{ 0xFF, 0xD8, 0xFE };
  try std.testing.expect(!isJpeg(&invalid));

  const short = [_]u8{ 0xFF, 0xD8 };
  try std.testing.expect(!isJpeg(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isJpeg(&empty));
}

test "is jpeg2000" {
  const jp2Buf = [_]u8{ 0x00, 0x00, 0x00, 0x0C, 0x6A, 0x50, 0x20, 0x20, 0x0D, 0x0A, 0x87, 0x0A, 0x00, 0x00 };
  try std.testing.expect(isJpeg2000(&jp2Buf));

  const invalid = [_]u8{ 0x00, 0x00, 0x00, 0x0C, 0x6A, 0x50, 0x20, 0xFF, 0x0D, 0x0A, 0x87, 0x0A, 0x00 };
  try std.testing.expect(!isJpeg2000(&invalid));

  const short = [_]u8{ 0x00, 0x00, 0x00, 0x0C, 0x6A, 0x50, 0x20, 0x20, 0x0D, 0x0A, 0x87, 0x0A };
  try std.testing.expect(!isJpeg2000(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isJpeg2000(&empty));
}

test "is png" {
  const pngBuf = [_]u8{ 0x89, 0x50, 0x4E, 0x47, 0x00 };
  try std.testing.expect(isPng(&pngBuf));

  const invalid = [_]u8{ 0x89, 0x50, 0x4E, 0xFF };
  try std.testing.expect(!isPng(&invalid));

  const short = [_]u8{ 0x89, 0x50, 0x4E };
  try std.testing.expect(!isPng(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isPng(&empty));
}

test "is gif" {
  const gifBuf = [_]u8{ 0x47, 0x49, 0x46, 0x00 };
  try std.testing.expect(isGif(&gifBuf));

  const invalid = [_]u8{ 0x47, 0x49, 0xFF };
  try std.testing.expect(!isGif(&invalid));

  const short = [_]u8{ 0x47, 0x49 };
  try std.testing.expect(!isGif(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isGif(&empty));
}

test "is webp" {
  var webpBuf = [_]u8{0} ** 12;
  webpBuf[8] = 0x57;
  webpBuf[9] = 0x45;
  webpBuf[10] = 0x42;
  webpBuf[11] = 0x50;
  try std.testing.expect(isWebp(&webpBuf));

  var invalid = webpBuf;
  invalid[8] = 0xFF;
  try std.testing.expect(!isWebp(&invalid));

  const short = [_]u8{ 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x57, 0x45, 0x42 };
  try std.testing.expect(!isWebp(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isWebp(&empty));
}

test "is cr2" {
  var cr2_iiBuf = [_]u8{0} ** 11;
  cr2_iiBuf[0] = 0x49;
  cr2_iiBuf[1] = 0x49;
  cr2_iiBuf[2] = 0x2A;
  cr2_iiBuf[3] = 0x00;
  cr2_iiBuf[8] = 0x43;
  cr2_iiBuf[9] = 0x52;
  cr2_iiBuf[10] = 0x02;
  try std.testing.expect(isCr2(&cr2_iiBuf));

  var cr2_mmBuf = [_]u8{0} ** 11;
  cr2_mmBuf[0] = 0x4D;
  cr2_mmBuf[1] = 0x4D;
  cr2_mmBuf[2] = 0x00;
  cr2_mmBuf[3] = 0x2A;
  cr2_mmBuf[8] = 0x43;
  cr2_mmBuf[9] = 0x52;
  cr2_mmBuf[10] = 0x02;
  try std.testing.expect(isCr2(&cr2_mmBuf));

  var invalid = cr2_iiBuf;
  invalid[8] = 0xFF;
  try std.testing.expect(!isCr2(&invalid));

  const short = [_]u8{ 0x49, 0x49, 0x2A, 0x00, 0x00, 0x00, 0x00, 0x00, 0x43, 0x52 };
  try std.testing.expect(!isCr2(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isCr2(&empty));
}

test "is tiff" {
  var tiff_iiBuf = [_]u8{0} ** 10;
  tiff_iiBuf[0] = 0x49;
  tiff_iiBuf[1] = 0x49;
  tiff_iiBuf[2] = 0x2A;
  tiff_iiBuf[3] = 0x00;
  try std.testing.expect(isTiff(&tiff_iiBuf));

  var tiff_mmBuf = [_]u8{0} ** 10;
  tiff_mmBuf[0] = 0x4D;
  tiff_mmBuf[1] = 0x4D;
  tiff_mmBuf[2] = 0x00;
  tiff_mmBuf[3] = 0x2A;
  try std.testing.expect(isTiff(&tiff_mmBuf));

  var cr2Buf = [_]u8{0} ** 11;
  cr2Buf[0] = 0x49;
  cr2Buf[1] = 0x49;
  cr2Buf[2] = 0x2A;
  cr2Buf[3] = 0x00;
  cr2Buf[8] = 0x43;
  cr2Buf[9] = 0x52;
  cr2Buf[10] = 0x02;
  try std.testing.expect(!isTiff(&cr2Buf));

  const short = [_]u8{ 0x49, 0x49, 0x2A, 0x00 };
  try std.testing.expect(!isTiff(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isTiff(&empty));
}

test "is bmp" {
  const bmpBuf = [_]u8{ 0x42, 0x4D, 0x00 };
  try std.testing.expect(isBmp(&bmpBuf));

  const invalid = [_]u8{ 0x42, 0xFF };
  try std.testing.expect(!isBmp(&invalid));

  const short = [_]u8{ 0x42 };
  try std.testing.expect(!isBmp(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isBmp(&empty));
}

test "is jxr" {
  const jxrBuf = [_]u8{ 0x49, 0x49, 0xBC, 0x00 };
  try std.testing.expect(isJxr(&jxrBuf));

  const invalid = [_]u8{ 0x49, 0x49, 0xFF };
  try std.testing.expect(!isJxr(&invalid));

  const short = [_]u8{ 0x49, 0x49 };
  try std.testing.expect(!isJxr(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isJxr(&empty));
}

test "is psd" {
  const psdBuf = [_]u8{ 0x38, 0x42, 0x50, 0x53, 0x00 };
  try std.testing.expect(isPsd(&psdBuf));

  const invalid = [_]u8{ 0x38, 0x42, 0x50, 0xFF };
  try std.testing.expect(!isPsd(&invalid));

  const short = [_]u8{ 0x38, 0x42, 0x50 };
  try std.testing.expect(!isPsd(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isPsd(&empty));
}

test "is ico" {
  const icoBuf = [_]u8{ 0x00, 0x00, 0x01, 0x00, 0x00 };
  try std.testing.expect(isIco(&icoBuf));

  const invalid = [_]u8{ 0x00, 0x00, 0xFF, 0x00 };
  try std.testing.expect(!isIco(&invalid));

  const short = [_]u8{ 0x00, 0x00, 0x01 };
  try std.testing.expect(!isIco(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isIco(&empty));
}

test "is jxl" {
  const jxl_ffBuf = [_]u8{ 0xFF, 0x0A, 0x00 };
  try std.testing.expect(isJxl(&jxl_ffBuf));

  const jxl_sigBuf = [_]u8{ 0x00, 0x00, 0x00, 0x0C, 0x4A, 0x58, 0x4C, 0x20, 0x0D, 0x0A, 0x87, 0x0A, 0x00 };
  try std.testing.expect(isJxl(&jxl_sigBuf));

  const invalid = [_]u8{ 0xFF, 0xFF, 0x00 };
  try std.testing.expect(!isJxl(&invalid));

  const short_ffBuf = [_]u8{ 0xFF };
  try std.testing.expect(!isJxl(&short_ffBuf));

  const short_sigBuf = [_]u8{ 0x00, 0x00, 0x00, 0x0C, 0x4A, 0x58, 0x4C, 0x20, 0x0D, 0x0A, 0x87 };
  try std.testing.expect(!isJxl(&short_sigBuf));

  const empty = [_]u8{};
  try std.testing.expect(!isJxl(&empty));
}

test "is heif" {
  var heifBuf = [_]u8{0} ** 16;
  heifBuf[0] = 0x00;
  heifBuf[1] = 0x00;
  heifBuf[2] = 0x00;
  heifBuf[3] = 0x10; // boxLen = 16
  heifBuf[4] = 'f';
  heifBuf[5] = 't';
  heifBuf[6] = 'y';
  heifBuf[7] = 'p';
  heifBuf[8] = 'h';
  heifBuf[9] = 'e';
  heifBuf[10] = 'i';
  heifBuf[11] = 'c';
  try std.testing.expect(isHeif(&heifBuf));

  var mif1Buf = [_]u8{0} ** 20;
  mif1Buf[0] = 0x00;
  mif1Buf[1] = 0x00;
  mif1Buf[2] = 0x00;
  mif1Buf[3] = 0x14; // boxLen = 20
  mif1Buf[4] = 'f';
  mif1Buf[5] = 't';
  mif1Buf[6] = 'y';
  mif1Buf[7] = 'p';
  mif1Buf[8] = 'm';
  mif1Buf[9] = 'i';
  mif1Buf[10] = 'f';
  mif1Buf[11] = '1';
  mif1Buf[16] = 'h';
  mif1Buf[17] = 'e';
  mif1Buf[18] = 'i';
  mif1Buf[19] = 'c';
  try std.testing.expect(isHeif(&mif1Buf));

  var invalid = heifBuf;
  invalid[8] = 'x';
  try std.testing.expect(!isHeif(&invalid));

  const short = [_]u8{ 0x00, 0x00, 0x00, 0x10, 'f', 't', 'y', 'p', 'h', 'e', 'i', 'c' };
  try std.testing.expect(!isHeif(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isHeif(&empty));
}

test "is avif" {
  var avifBuf = [_]u8{0} ** 16;
  avifBuf[0] = 0x00;
  avifBuf[1] = 0x00;
  avifBuf[2] = 0x00;
  avifBuf[3] = 0x10; // boxLen = 16
  avifBuf[4] = 'f';
  avifBuf[5] = 't';
  avifBuf[6] = 'y';
  avifBuf[7] = 'p';
  avifBuf[8] = 'a';
  avifBuf[9] = 'v';
  avifBuf[10] = 'i';
  avifBuf[11] = 'f';
  try std.testing.expect(isAvif(&avifBuf));

  var compatBuf = [_]u8{0} ** 20;
  compatBuf[0] = 0x00;
  compatBuf[1] = 0x00;
  compatBuf[2] = 0x00;
  compatBuf[3] = 0x14; // boxLen = 20
  compatBuf[4] = 'f';
  compatBuf[5] = 't';
  compatBuf[6] = 'y';
  compatBuf[7] = 'p';
  compatBuf[8] = 'm';
  compatBuf[9] = 'i';
  compatBuf[10] = 'f';
  compatBuf[11] = '1';
  compatBuf[16] = 'a';
  compatBuf[17] = 'v';
  compatBuf[18] = 'i';
  compatBuf[19] = 'f';
  try std.testing.expect(isAvif(&compatBuf));

  var invalid = avifBuf;
  invalid[8] = 'x';
  try std.testing.expect(!isAvif(&invalid));

  const short = [_]u8{ 0x00, 0x00, 0x00, 0x10, 'f', 't', 'y', 'p', 'a', 'v', 'i', 'f' };
  try std.testing.expect(!isAvif(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isAvif(&empty));
}

test "is ora" {
  var oraBuf = [_]u8{0} ** 58;
  oraBuf[0] = 0x50;
  oraBuf[1] = 0x4B;
  oraBuf[2] = 0x03;
  oraBuf[3] = 0x04;
  const mimetype = "mimetypeimage/openraster";
  for (mimetype, 0..) |c, i| {
      oraBuf[30 + i] = c;
  }
  try std.testing.expect(isOra(&oraBuf));

  var invalid = oraBuf;
  invalid[30] = 0xFF;
  try std.testing.expect(!isOra(&invalid));

  const short = [_]u8{ 0x50, 0x4B, 0x03, 0x04 };
  try std.testing.expect(!isOra(&short));

  const empty = [_]u8{};
  try std.testing.expect(!isOra(&empty));
}

test "is djvu" {
  var djvuBuf = [_]u8{0} ** 15;
  djvuBuf[0] = 0x41;
  djvuBuf[1] = 0x54;
  djvuBuf[2] = 0x26;
  djvuBuf[3] = 0x54;
  djvuBuf[4] = 0x46;
  djvuBuf[5] = 0x4F;
  djvuBuf[6] = 0x52;
  djvuBuf[7] = 0x4D;
  djvuBuf[12] = 0x44;
  djvuBuf[13] = 0x4A;
  djvuBuf[14] = 0x56;
  try std.testing.expect(is_djvu(&djvuBuf));

  var invalid = djvuBuf;
  invalid[12] = 0xFF;
  try std.testing.expect(!is_djvu(&invalid));

  const short = [_]u8{ 0x41, 0x54, 0x26, 0x54, 0x46, 0x4F, 0x52, 0x4D };
  try std.testing.expect(!is_djvu(&short));

  const empty = [_]u8{};
  try std.testing.expect(!is_djvu(&empty));
}