// zig fmt: off
const std = @import("std");
const m = @import("cmp.zig");

const DocType = enum {
  Text,
  Spreadsheet,
  Presentation,
};

/// 与えられたバッファがOpenDocument Textファイルであるかどうかを返します。
pub fn isOdt(buf: []const u8) bool {
  return odf(buf) == .Text;
}

/// 与えられたバッファがOpenDocument Spreadsheetファイルであるかどうかを返します。
pub fn isOds(buf: []const u8) bool {
  return odf(buf) == .Spreadsheet;
}

/// 与えられたバッファがOpenDocument Presentationファイルであるかどうかを返します。
pub fn isOdp(buf: []const u8) bool {
  return odf(buf) == .Presentation;
}

fn odf(buf: []const u8) ?DocType {
  const signature = [_]u8{ 'P', 'K', 0x03, 0x04 };
  if (!m.compareBytes(buf, &signature, 0)) return null;

  const mimetype = [_]u8{ 'm', 'i', 'm', 'e', 't', 'y', 'p', 'e' };
  if (!m.compareBytes(buf, &mimetype, 0x1E)) return null;

  const text_mime = [_]u8{
    'v', 'n', 'd', '.', 'o', 'a', 's', 'i', 's', '.',
    'o', 'p', 'e', 'n', 'd', 'o', 'c', 'u', 'm', 'e', 'n', 't', '.',
    't', 'e', 'x', 't'
  };
  if (m.compareBytes(buf, &text_mime, 0x32)) return .Text;

  const spreadsheet_mime = [_]u8{
    'v', 'n', 'd', '.', 'o', 'a', 's', 'i', 's', '.',
    'o', 'p', 'e', 'n', 'd', 'o', 'c', 'u', 'm', 'e', 'n', 't', '.',
    's', 'p', 'r', 'e', 'a', 'd', 's', 'h', 'e', 'e', 't'
  };
  if (m.compareBytes(buf, &spreadsheet_mime, 0x32)) return .Spreadsheet;

  const presentation_mime = [_]u8{
    'v', 'n', 'd', '.', 'o', 'a', 's', 'i', 's', '.',
    'o', 'p', 'e', 'n', 'd', 'o', 'c', 'u', 'm', 'e', 'n', 't', '.',
    'p', 'r', 'e', 's', 'e', 'n', 't', 'a', 't', 'i', 'o', 'n'
  };
  if (m.compareBytes(buf, &presentation_mime, 0x32)) return .Presentation;

  return null;
}

test "is odt" {
  var odtBuf = [_]u8{0} ** 100;
  odtBuf[0] = 'P';
  odtBuf[1] = 'K';
  odtBuf[2] = 0x03;
  odtBuf[3] = 0x04;
  const mimetype = "mimetype";
  for (mimetype, 0..) |c, i| {
    odtBuf[0x1E + i] = c;
  }
  const text_mime = "vnd.oasis.opendocument.text";
  for (text_mime, 0..) |c, i| {
    odtBuf[0x32 + i] = c;
  }
  try std.testing.expect(isOdt(&odtBuf));
  try std.testing.expect(!isOds(&odtBuf));
  try std.testing.expect(!isOdp(&odtBuf));

  var invalid = odtBuf;
  invalid[0] = 'X';
  try std.testing.expect(!isOdt(&invalid));

  const short = [_]u8{ 'P', 'K' };
  try std.testing.expect(!isOdt(&short));
}

test "is ods" {
  var odsBuf = [_]u8{0} ** 100;
  odsBuf[0] = 'P';
  odsBuf[1] = 'K';
  odsBuf[2] = 0x03;
  odsBuf[3] = 0x04;
  const mimetype = "mimetype";
  for (mimetype, 0..) |c, i| {
    odsBuf[0x1E + i] = c;
  }
  const spreadsheetMime = "vnd.oasis.opendocument.spreadsheet";
  for (spreadsheetMime, 0..) |c, i| {
    odsBuf[0x32 + i] = c;
  }
  try std.testing.expect(isOds(&odsBuf));
  try std.testing.expect(!isOdt(&odsBuf));
  try std.testing.expect(!isOdp(&odsBuf));

  var invalid = odsBuf;
  invalid[0x1E] = 'x';
  try std.testing.expect(!isOds(&invalid));
}

test "is odp" {
  var odpBuf = [_]u8{0} ** 100;
  odpBuf[0] = 'P';
  odpBuf[1] = 'K';
  odpBuf[2] = 0x03;
  odpBuf[3] = 0x04;
  const mimetype = "mimetype";
  for (mimetype, 0..) |c, i| {
    odpBuf[0x1E + i] = c;
  }
  const presentationMime = "vnd.oasis.opendocument.presentation";
  for (presentationMime, 0..) |c, i| {
    odpBuf[0x32 + i] = c;
  }
  try std.testing.expect(isOdp(&odpBuf));
  try std.testing.expect(!isOdt(&odpBuf));
  try std.testing.expect(!isOds(&odpBuf));

  var invalid = odpBuf;
  invalid[0x32] = 'x';
  try std.testing.expect(!isOdp(&invalid));
}