// zig fmt: off
const std = @import("std");

/// 与えられたバッファがHTMLデータであるかどうかを返します。
/// [WHATWG](https://mimesniff.spec.whatwg.org/) 仕様に準拠しています。
pub fn isHtml(buf: []const u8) bool {
  const values = [_][]const u8{
    "<!DOCTYPE HTML",
    "<HTML",
    "<HEAD",
    "<SCRIPT",
    "<IFRAME",
    "<H1",
    "<DIV",
    "<FONT",
    "<TABLE",
    "<A",
    "<STYLE",
    "<TITLE",
    "<B",
    "<BODY",
    "<BR",
    "<P",
    "<!--",
  };
  const trimmed_buf = trimStartWhitespaces(buf);

  for (values) |val| {
    if (startsWithIgnoreAsciiCase(trimmed_buf, val) and trimmed_buf.len > val.len) {
      const next_byte = trimmed_buf[val.len];
      if (next_byte == 0x20 or next_byte == 0x3E) {
        return true;
      }
    }
  }

  return false;
}

/// 与えられたバッファがXMLデータであるかどうかを返します。
/// [WHATWG](https://mimesniff.spec.whatwg.org/) 仕様に準拠しています。
pub fn isXml(buf: []const u8) bool {
  const val = "<?xml";
  var trimmed_buf = trimStartWhitespaces(buf);
  trimmed_buf = trimStartByteOrderMarks(trimmed_buf);
  return startsWithIgnoreAsciiCase(trimmed_buf, val);
}

/// バッファの先頭にある空白を削除します。
/// https://mimesniff.spec.whatwg.org の空白の定義に従います。
fn trimStartWhitespaces(buf: []const u8) []const u8 {
  var i: usize = 0;
  while (i < buf.len) : (i += 1) {
    switch (buf[i]) {
      0x09, 0x0A, 0x0C, 0x0D, 0x20 => continue,
      else => break,
    }
  }
  return buf[i..];
}

/// バッファの先頭にあるバイトオーダーマーク（BOM）を削除します。
fn trimStartByteOrderMarks(buf: []const u8) []const u8 {
  if (buf.len >= 3 and buf[0] == 0xEF and buf[1] == 0xBB and buf[2] == 0xBF) {
    return buf[3..]; // UTF-8 BOM
  }
  if (buf.len >= 2 and ((buf[0] == 0xFE and buf[1] == 0xFF) or (buf[0] == 0xFF and buf[1] == 0xFE))) {
    return buf[2..]; // UTF-16 BE or LE BOM
  }
  return buf;
}

/// ASCIIの大文字小文字を無視して、バッファが指定された文字列で始まるかどうかを確認します。
fn startsWithIgnoreAsciiCase(buf: []const u8, needle: []const u8) bool {
  if (buf.len < needle.len) return false;

  for (buf[0..needle.len], needle) |b, n| {
    const b_lower = if (b >= 'A' and b <= 'Z') b + 32 else b;
    const n_lower = if (n >= 'A' and n <= 'Z') n + 32 else n;
    if (b_lower != n_lower) return false;
  }
  return true;
}

/// 与えられたバッファがシェルスクリプトであるかどうかを返します。
pub fn isShellscript(buf: []const u8) bool {
  return buf.len > 1 and buf[0] == '#' and buf[1] == '!';
}

test "trim start whitespaces" {
  const input = [_]u8{ 0x09, 0x0A, 0x0C, 0x0D, 0x20, 'A', 'B', 'C' };
  try std.testing.expectEqualSlices(u8, "ABC", trimStartWhitespaces(&input));

  const no_ws = "abc";
  try std.testing.expectEqualSlices(u8, "abc", trimStartWhitespaces(no_ws));

  const empty = [_]u8{};
  try std.testing.expectEqualSlices(u8, &[_]u8{}, trimStartWhitespaces(&empty));
}

test "trim start byte order marks" {
  const utf8_bom = [_]u8{ 0xEF, 0xBB, 0xBF, 'A' };
  try std.testing.expectEqualSlices(u8, "A", trimStartByteOrderMarks(&utf8_bom));

  const utf16_be = [_]u8{ 0xFE, 0xFF, 'A' };
  try std.testing.expectEqualSlices(u8, "A", trimStartByteOrderMarks(&utf16_be));

  const utf16_le = [_]u8{ 0xFF, 0xFE, 'A' };
  try std.testing.expectEqualSlices(u8, "A", trimStartByteOrderMarks(&utf16_le));

  const no_bom = "abc";
  try std.testing.expectEqualSlices(u8, "abc", trimStartByteOrderMarks(no_bom));
}

test "is html" {
  try std.testing.expect(isHtml("<HTML "));
  try std.testing.expect(isHtml("   <BODY>"));
  try std.testing.expect(isHtml("<!DOCTYPE HTML>"));
  try std.testing.expect(isHtml("<H1>"));

  try std.testing.expect(!isHtml("<"));
  try std.testing.expect(!isHtml("<HTML"));
  try std.testing.expect(!isHtml("abc"));

  try std.testing.expect(!isHtml(""));
}

test "is xml" {
  try std.testing.expect(isXml("<?xml version=\"1.0\"?>"));
  try std.testing.expect(isXml("   <?XML >"));
  try std.testing.expect(isXml("\xEF\xBB\xBF<?xml>")); // UTF-8 BOM

  try std.testing.expect(!isXml("<xml>"));
  try std.testing.expect(!isXml("abc"));

  try std.testing.expect(!isXml(""));
}

test "is shellscript" {
  try std.testing.expect(isShellscript("#!/bin/sh"));
  try std.testing.expect(isShellscript("#! "));

  try std.testing.expect(!isShellscript("# "));
  try std.testing.expect(!isShellscript("abc"));

  try std.testing.expect(!isShellscript("#"));
}